
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import migrate
import sqlalchemy
from sqlalchemy import (func, Index, inspect, orm, String, Table, type_coerce)


# The _upgrade...get_duplicate() def's are separate functions to
# accommodate sqlite which locks the database against updates as long as
# db_recs is active.
# In addition, sqlite doesn't support the function 'concat' between
# Strings and Integers, so, the updating of records is also adjusted.
def _upgrade_metadef_namespaces_get_duplicates(migrate_engine):
    meta = sqlalchemy.schema.MetaData(migrate_engine)
    metadef_namespaces = Table('metadef_namespaces', meta, autoload=True)

    session = orm.sessionmaker(bind=migrate_engine)()
    db_recs = (session.query(func.min(metadef_namespaces.c.id),
                             metadef_namespaces.c.namespace)
               .group_by(metadef_namespaces.c.namespace)
               .having(func.count(metadef_namespaces.c.namespace) > 1))
    dbrecs = []
    for row in db_recs:
        dbrecs.append({'id': row[0], 'namespace': row[1]})
    session.close()

    return dbrecs


def _upgrade_metadef_objects_get_duplicates(migrate_engine):
    meta = sqlalchemy.schema.MetaData(migrate_engine)
    metadef_objects = Table('metadef_objects', meta, autoload=True)

    session = orm.sessionmaker(bind=migrate_engine)()
    db_recs = (session.query(func.min(metadef_objects.c.id),
                             metadef_objects.c.namespace_id,
                             metadef_objects.c.name)
               .group_by(metadef_objects.c.namespace_id,
                         metadef_objects.c.name)
               .having(func.count() > 1))
    dbrecs = []
    for row in db_recs:
        dbrecs.append({'id': row[0], 'namespace_id': row[1], 'name': row[2]})
    session.close()

    return dbrecs


def _upgrade_metadef_properties_get_duplicates(migrate_engine):
    meta = sqlalchemy.schema.MetaData(migrate_engine)
    metadef_properties = Table('metadef_properties', meta, autoload=True)

    session = orm.sessionmaker(bind=migrate_engine)()
    db_recs = (session.query(func.min(metadef_properties.c.id),
                             metadef_properties.c.namespace_id,
                             metadef_properties.c.name)
               .group_by(metadef_properties.c.namespace_id,
                         metadef_properties.c.name)
               .having(func.count() > 1))
    dbrecs = []
    for row in db_recs:
        dbrecs.append({'id': row[0], 'namespace_id': row[1], 'name': row[2]})
    session.close()

    return dbrecs


def _upgrade_metadef_tags_get_duplicates(migrate_engine):
    meta = sqlalchemy.schema.MetaData(migrate_engine)
    metadef_tags = Table('metadef_tags', meta, autoload=True)

    session = orm.sessionmaker(bind=migrate_engine)()
    db_recs = (session.query(func.min(metadef_tags.c.id),
                             metadef_tags.c.namespace_id,
                             metadef_tags.c.name)
               .group_by(metadef_tags.c.namespace_id,
                         metadef_tags.c.name)
               .having(func.count() > 1))
    dbrecs = []
    for row in db_recs:
        dbrecs.append({'id': row[0], 'namespace_id': row[1], 'name': row[2]})
    session.close()

    return dbrecs


def _upgrade_metadef_resource_types_get_duplicates(migrate_engine):
    meta = sqlalchemy.schema.MetaData(migrate_engine)
    metadef_resource_types = Table('metadef_resource_types', meta,
                                   autoload=True)

    session = orm.sessionmaker(bind=migrate_engine)()
    db_recs = (session.query(func.min(metadef_resource_types.c.id),
                             metadef_resource_types.c.name)
               .group_by(metadef_resource_types.c.name)
               .having(func.count(metadef_resource_types.c.name) > 1))
    dbrecs = []
    for row in db_recs:
        dbrecs.append({'id': row[0], 'name': row[1]})
    session.close()

    return dbrecs


def _upgrade_data(migrate_engine):
    # Rename duplicates to be unique.
    meta = sqlalchemy.schema.MetaData(migrate_engine)

    # ORM tables
    metadef_namespaces = Table('metadef_namespaces', meta, autoload=True)
    metadef_objects = Table('metadef_objects', meta, autoload=True)
    metadef_properties = Table('metadef_properties', meta, autoload=True)
    metadef_tags = Table('metadef_tags', meta, autoload=True)
    metadef_resource_types = Table('metadef_resource_types', meta,
                                   autoload=True)

    # Fix duplicate metadef_namespaces
    # Update the non-first record(s) with an unique namespace value
    dbrecs = _upgrade_metadef_namespaces_get_duplicates(migrate_engine)
    for row in dbrecs:
        s = (metadef_namespaces.update()
             .where(metadef_namespaces.c.id > row['id'])
             .where(metadef_namespaces.c.namespace == row['namespace'])
             )
        if migrate_engine.name == 'sqlite':
            s = (s.values(namespace=(row['namespace'] + '-DUPL-' +
                                     type_coerce(metadef_namespaces.c.id,
                                                 String)),
                          display_name=(row['namespace'] + '-DUPL-' +
                                        type_coerce(metadef_namespaces.c.id,
                                                    String))))
        else:
            s = s.values(namespace=func.concat(row['namespace'],
                                               '-DUPL-',
                                               metadef_namespaces.c.id),
                         display_name=func.concat(row['namespace'],
                                                  '-DUPL-',
                                                  metadef_namespaces.c.id))
        s.execute()

    # Fix duplicate metadef_objects
    dbrecs = _upgrade_metadef_objects_get_duplicates(migrate_engine)
    for row in dbrecs:
        s = (metadef_objects.update()
             .where(metadef_objects.c.id > row['id'])
             .where(metadef_objects.c.namespace_id == row['namespace_id'])
             .where(metadef_objects.c.name == str(row['name']))
             )
        if migrate_engine.name == 'sqlite':
            s = (s.values(name=(row['name'] + '-DUPL-'
                          + type_coerce(metadef_objects.c.id, String))))
        else:
            s = s.values(name=func.concat(row['name'], '-DUPL-',
                                          metadef_objects.c.id))
        s.execute()

    # Fix duplicate metadef_properties
    dbrecs = _upgrade_metadef_properties_get_duplicates(migrate_engine)
    for row in dbrecs:
        s = (metadef_properties.update()
             .where(metadef_properties.c.id > row['id'])
             .where(metadef_properties.c.namespace_id == row['namespace_id'])
             .where(metadef_properties.c.name == str(row['name']))
             )
        if migrate_engine.name == 'sqlite':
            s = (s.values(name=(row['name'] + '-DUPL-' +
                                type_coerce(metadef_properties.c.id, String)))
                 )
        else:
            s = s.values(name=func.concat(row['name'], '-DUPL-',
                                          metadef_properties.c.id))
        s.execute()

    # Fix duplicate metadef_tags
    dbrecs = _upgrade_metadef_tags_get_duplicates(migrate_engine)
    for row in dbrecs:
        s = (metadef_tags.update()
             .where(metadef_tags.c.id > row['id'])
             .where(metadef_tags.c.namespace_id == row['namespace_id'])
             .where(metadef_tags.c.name == str(row['name']))
             )
        if migrate_engine.name == 'sqlite':
            s = (s.values(name=(row['name'] + '-DUPL-' +
                                type_coerce(metadef_tags.c.id, String)))
                 )
        else:
            s = s.values(name=func.concat(row['name'], '-DUPL-',
                                          metadef_tags.c.id))
        s.execute()

    # Fix duplicate metadef_resource_types
    dbrecs = _upgrade_metadef_resource_types_get_duplicates(migrate_engine)
    for row in dbrecs:
        s = (metadef_resource_types.update()
             .where(metadef_resource_types.c.id > row['id'])
             .where(metadef_resource_types.c.name == str(row['name']))
             )
        if migrate_engine.name == 'sqlite':
            s = (s.values(name=(row['name'] + '-DUPL-' +
                                type_coerce(metadef_resource_types.c.id,
                                            String)))
                 )
        else:
            s = s.values(name=func.concat(row['name'], '-DUPL-',
                                          metadef_resource_types.c.id))
        s.execute()


def _update_sqlite_namespace_id_name_constraint(metadef, metadef_namespaces,
                                                new_constraint_name,
                                                new_fk_name):
    migrate.UniqueConstraint(
        metadef.c.namespace_id, metadef.c.name).drop()
    migrate.UniqueConstraint(
        metadef.c.namespace_id, metadef.c.name,
        name=new_constraint_name).create()
    migrate.ForeignKeyConstraint(
        [metadef.c.namespace_id],
        [metadef_namespaces.c.id],
        name=new_fk_name).create()


def _drop_unique_constraint_if_exists(inspector, table_name, metadef):
    name = _get_unique_constraint_name(inspector,
                                       table_name,
                                       ['namespace_id', 'name'])
    if name:
        migrate.UniqueConstraint(metadef.c.namespace_id,
                                 metadef.c.name,
                                 name=name).drop()


def _drop_index_with_fk_constraint(metadef, metadef_namespaces,
                                   index_name,
                                   fk_old_name, fk_new_name):

    fkc = migrate.ForeignKeyConstraint([metadef.c.namespace_id],
                                       [metadef_namespaces.c.id],
                                       name=fk_old_name)
    fkc.drop()

    if index_name:
        Index(index_name, metadef.c.namespace_id).drop()

    # Rename the fk for consistency across all db's
    fkc = migrate.ForeignKeyConstraint([metadef.c.namespace_id],
                                       [metadef_namespaces.c.id],
                                       name=fk_new_name)
    fkc.create()


def _get_unique_constraint_name(inspector, table_name, columns):
    constraints = inspector.get_unique_constraints(table_name)
    for constraint in constraints:
        if set(constraint['column_names']) == set(columns):
            return constraint['name']
    return None


def _get_fk_constraint_name(inspector, table_name, columns):
    constraints = inspector.get_foreign_keys(table_name)
    for constraint in constraints:
        if set(constraint['constrained_columns']) == set(columns):
            return constraint['name']
    return None


def upgrade(migrate_engine):

    _upgrade_data(migrate_engine)

    meta = sqlalchemy.MetaData()
    meta.bind = migrate_engine
    inspector = inspect(migrate_engine)

    # ORM tables
    metadef_namespaces = Table('metadef_namespaces', meta, autoload=True)
    metadef_objects = Table('metadef_objects', meta, autoload=True)
    metadef_properties = Table('metadef_properties', meta, autoload=True)
    metadef_tags = Table('metadef_tags', meta, autoload=True)
    metadef_ns_res_types = Table('metadef_namespace_resource_types',
                                 meta, autoload=True)
    metadef_resource_types = Table('metadef_resource_types', meta,
                                   autoload=True)

    # Drop the bad, non-unique indices.
    if migrate_engine.name == 'sqlite':
        # For sqlite:
        # Only after the unique constraints have been added should the indices
        # be dropped. If done the other way, sqlite complains during
        # constraint adding/dropping that the index does/does not exist.
        # Note: The _get_unique_constraint_name, _get_fk_constraint_name
        # return None for constraints that do in fact exist. Also,
        # get_index_names returns names, but, the names can not be used with
        # the Index(name, blah).drop() command, so, putting sqlite into
        # it's own section.

        # Objects
        _update_sqlite_namespace_id_name_constraint(
            metadef_objects, metadef_namespaces,
            'uq_metadef_objects_namespace_id_name',
            'metadef_objects_fk_1')

        # Properties
        _update_sqlite_namespace_id_name_constraint(
            metadef_properties, metadef_namespaces,
            'uq_metadef_properties_namespace_id_name',
            'metadef_properties_fk_1')

        # Tags
        _update_sqlite_namespace_id_name_constraint(
            metadef_tags, metadef_namespaces,
            'uq_metadef_tags_namespace_id_name',
            'metadef_tags_fk_1')

        # Namespaces
        migrate.UniqueConstraint(
            metadef_namespaces.c.namespace).drop()
        migrate.UniqueConstraint(
            metadef_namespaces.c.namespace,
            name='uq_metadef_namespaces_namespace').create()

        # ResourceTypes
        migrate.UniqueConstraint(
            metadef_resource_types.c.name).drop()
        migrate.UniqueConstraint(
            metadef_resource_types.c.name,
            name='uq_metadef_resource_types_name').create()

        # Now drop the bad indices
        Index('ix_metadef_objects_namespace_id',
              metadef_objects.c.namespace_id,
              metadef_objects.c.name).drop()
        Index('ix_metadef_properties_namespace_id',
              metadef_properties.c.namespace_id,
              metadef_properties.c.name).drop()
        Index('ix_metadef_tags_namespace_id',
              metadef_tags.c.namespace_id,
              metadef_tags.c.name).drop()
    else:
        # First drop the bad non-unique indices.
        # To do that (for mysql), must first drop foreign key constraints
        # BY NAME and then drop the bad indices.
        # Finally, re-create the foreign key constraints with a consistent
        # name.

        # DB2 still has unique constraints, but, they are badly named.
        # Drop them, they will be recreated at the final step.
        name = _get_unique_constraint_name(inspector, 'metadef_namespaces',
                                           ['namespace'])
        if name:
            migrate.UniqueConstraint(metadef_namespaces.c.namespace,
                                     name=name).drop()
        _drop_unique_constraint_if_exists(inspector, 'metadef_objects',
                                          metadef_objects)
        _drop_unique_constraint_if_exists(inspector, 'metadef_properties',
                                          metadef_properties)
        _drop_unique_constraint_if_exists(inspector, 'metadef_tags',
                                          metadef_tags)
        name = _get_unique_constraint_name(inspector, 'metadef_resource_types',
                                           ['name'])
        if name:
            migrate.UniqueConstraint(metadef_resource_types.c.name,
                                     name=name).drop()

        # Objects
        _drop_index_with_fk_constraint(
            metadef_objects, metadef_namespaces,
            'ix_metadef_objects_namespace_id',
            _get_fk_constraint_name(
                inspector, 'metadef_objects', ['namespace_id']),
            'metadef_objects_fk_1')

        # Properties
        _drop_index_with_fk_constraint(
            metadef_properties, metadef_namespaces,
            'ix_metadef_properties_namespace_id',
            _get_fk_constraint_name(
                inspector, 'metadef_properties', ['namespace_id']),
            'metadef_properties_fk_1')

        # Tags
        _drop_index_with_fk_constraint(
            metadef_tags, metadef_namespaces,
            'ix_metadef_tags_namespace_id',
            _get_fk_constraint_name(
                inspector, 'metadef_tags', ['namespace_id']),
            'metadef_tags_fk_1')

    # Drop Others without fk constraints.
    Index('ix_metadef_namespaces_namespace',
          metadef_namespaces.c.namespace).drop()

    # The next two don't exist in ibm_db_sa, but, drop them everywhere else.
    if migrate_engine.name != 'ibm_db_sa':
        Index('ix_metadef_resource_types_name',
              metadef_resource_types.c.name).drop()
        # Not needed due to primary key on same columns
        Index('ix_metadef_ns_res_types_res_type_id_ns_id',
              metadef_ns_res_types.c.resource_type_id,
              metadef_ns_res_types.c.namespace_id).drop()

    # Now, add back the dropped indexes as unique constraints
    if migrate_engine.name != 'sqlite':
        # Namespaces
        migrate.UniqueConstraint(
            metadef_namespaces.c.namespace,
            name='uq_metadef_namespaces_namespace').create()

        # Objects
        migrate.UniqueConstraint(
            metadef_objects.c.namespace_id,
            metadef_objects.c.name,
            name='uq_metadef_objects_namespace_id_name').create()

        # Properties
        migrate.UniqueConstraint(
            metadef_properties.c.namespace_id,
            metadef_properties.c.name,
            name='uq_metadef_properties_namespace_id_name').create()

        # Tags
        migrate.UniqueConstraint(
            metadef_tags.c.namespace_id,
            metadef_tags.c.name,
            name='uq_metadef_tags_namespace_id_name').create()

        # Resource Types
        migrate.UniqueConstraint(
            metadef_resource_types.c.name,
            name='uq_metadef_resource_types_name').create()



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

__all__ = [
    'run',
]

from oslo_concurrency import lockutils
from oslo_log import log as logging
from oslo_utils import encodeutils
from oslo_utils import excutils
import six

from glance.api.v2 import images as v2_api
from glance.common import exception
from glance.common.scripts import utils as script_utils
from glance.common import store_utils
from glance.i18n import _, _LE, _LI, _LW

LOG = logging.getLogger(__name__)


def run(t_id, context, task_repo, image_repo, image_factory):
    LOG.info(_LI('Task %(task_id)s beginning import '
                 'execution.'), {'task_id': t_id})
    _execute(t_id, task_repo, image_repo, image_factory)


# NOTE(nikhil): This lock prevents more than N number of threads to be spawn
# simultaneously. The number N represents the number of threads in the
# executor pool. The value is set to 10 in the eventlet executor.
@lockutils.synchronized("glance_import")
def _execute(t_id, task_repo, image_repo, image_factory):
    task = script_utils.get_task(task_repo, t_id)

    if task is None:
        # NOTE: This happens if task is not found in the database. In
        # such cases, there is no way to update the task status so,
        # it's ignored here.
        return

    try:
        task_input = script_utils.unpack_task_input(task)

        uri = script_utils.validate_location_uri(task_input.get('import_from'))
        image_id = import_image(image_repo, image_factory, task_input, t_id,
                                uri)

        task.succeed({'image_id': image_id})
    except Exception as e:
        # Note: The message string contains Error in it to indicate
        # in the task.message that it's a error message for the user.

        # TODO(nikhil): need to bring back save_and_reraise_exception when
        # necessary
        err_msg = ("Error: " + six.text_type(type(e)) + ': ' +
                   encodeutils.exception_to_unicode(e))
        log_msg = _LE(err_msg + ("Task ID %s" % task.task_id))  # noqa
        LOG.exception(log_msg)

        task.fail(_LE(err_msg))  # noqa
    finally:
        task_repo.save(task)


def import_image(image_repo, image_factory, task_input, task_id, uri):
    original_image = create_image(image_repo, image_factory,
                                  task_input.get('image_properties'), task_id)
    # NOTE: set image status to saving just before setting data
    original_image.status = 'saving'
    image_repo.save(original_image)
    image_id = original_image.image_id

    # NOTE: Retrieving image from the database because the Image object
    # returned from create_image method does not have appropriate factories
    # wrapped around it.
    new_image = image_repo.get(image_id)
    set_image_data(new_image, uri, task_id)

    try:
        # NOTE: Check if the Image is not deleted after setting the data
        # before saving the active image. Here if image status is
        # saving, then new_image is saved as it contains updated location,
        # size, virtual_size and checksum information and the status of
        # new_image is already set to active in set_image_data() call.
        image = image_repo.get(image_id)
        if image.status == 'saving':
            image_repo.save(new_image)
            return image_id
        else:
            msg = _("The Image %(image_id)s object being created by this task "
                    "%(task_id)s, is no longer in valid status for further "
                    "processing.") % {"image_id": image_id,
                                      "task_id": task_id}
            raise exception.Conflict(msg)
    except (exception.Conflict, exception.NotFound,
            exception.NotAuthenticated):
        with excutils.save_and_reraise_exception():
            if new_image.locations:
                for location in new_image.locations:
                    store_utils.delete_image_location_from_backend(
                        new_image.context,
                        image_id,
                        location)


def create_image(image_repo, image_factory, image_properties, task_id):
    properties = {}
    # NOTE: get the base properties
    for key in v2_api.get_base_properties():
        try:
            properties[key] = image_properties.pop(key)
        except KeyError:
            LOG.debug("Task ID %(task_id)s: Ignoring property %(k)s for "
                      "setting base properties while creating "
                      "Image.", {'task_id': task_id, 'k': key})

    # NOTE: get the rest of the properties and pass them as
    # extra_properties for Image to be created with them.
    properties['extra_properties'] = image_properties
    script_utils.set_base_image_properties(properties=properties)

    image = image_factory.new_image(**properties)
    image_repo.add(image)
    return image


def set_image_data(image, uri, task_id, backend=None, set_active=True,
                   callback=None):
    data_iter = None
    try:
        LOG.info(_LI("Task %(task_id)s: Got image data uri %(data_uri)s to be "
                 "imported"), {"data_uri": uri, "task_id": task_id})
        data_iter = script_utils.get_image_data_iter(uri)
        if callback:
            # If a callback was provided, wrap our data iterator to call
            # the function every 60 seconds.
            data_iter = script_utils.CallbackIterator(
                data_iter, callback, min_interval=60)
        image.set_data(data_iter, backend=backend, set_active=set_active)
    except Exception as e:
        with excutils.save_and_reraise_exception():
            LOG.warn(_LW("Task %(task_id)s failed with exception %(error)s") %
                     {"error": encodeutils.exception_to_unicode(e),
                      "task_id": task_id})
            LOG.info(_LI("Task %(task_id)s: Could not import image file"
                         " %(image_data)s"), {"image_data": uri,
                                              "task_id": task_id})
    finally:
        if hasattr(data_iter, 'close'):
            data_iter.close()



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
# Copyright 2013 OpenStack Foundation
# Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Database setup and migration commands."""

import os
import threading

from oslo_config import cfg
from oslo_db import options as db_options
from stevedore import driver


_IMPL = None
_LOCK = threading.Lock()

db_options.set_defaults(cfg.CONF)


def get_backend():
    global _IMPL
    if _IMPL is None:
        with _LOCK:
            if _IMPL is None:
                _IMPL = driver.DriverManager(
                    "glance.database.migration_backend",
                    cfg.CONF.database.backend).driver
    return _IMPL


# Migration-related constants
EXPAND_BRANCH = 'expand'
CONTRACT_BRANCH = 'contract'
CURRENT_RELEASE = 'wallaby'
ALEMBIC_INIT_VERSION = 'liberty'
LATEST_REVISION = 'wallaby_contract01'
INIT_VERSION = 0

MIGRATE_REPO_PATH = os.path.join(
    os.path.abspath(os.path.dirname(__file__)),
    'sqlalchemy',
    'migrate_repo',
)



#    Copyright (c) 2016 Hewlett Packard Enterprise Software, LLC
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from sqlalchemy import MetaData, Table


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine

    resource_types_table = Table('metadef_resource_types', meta, autoload=True)

    resource_types_table.update(values={'name': 'OS::Nova::Server'}).where(
        resource_types_table.c.name == 'OS::Nova::Instance').execute()



# -*- coding: utf-8 -*-
# Copyright 2020, Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from glance.api import common
import glance.async_
from glance.common import exception
from glance.common import wsgi_app
from glance.tests import utils as test_utils


class TestWsgiAppInit(test_utils.BaseTestCase):
    @mock.patch('glance.common.config.load_paste_app')
    @mock.patch('glance.async_.set_threadpool_model')
    @mock.patch('glance.common.wsgi_app._get_config_files')
    def test_wsgi_init_sets_thread_settings(self, mock_config_files,
                                            mock_set_model,
                                            mock_load):
        mock_config_files.return_value = []
        self.config(task_pool_threads=123, group='wsgi')
        common.DEFAULT_POOL_SIZE = 1024
        wsgi_app.init_app()
        # Make sure we declared the system threadpool model as native
        mock_set_model.assert_called_once_with('native')
        # Make sure we set the default pool size
        self.assertEqual(123, common.DEFAULT_POOL_SIZE)
        mock_load.assert_called_once_with('glance-api')

    @mock.patch('atexit.register')
    @mock.patch('glance.common.config.load_paste_app')
    @mock.patch('glance.async_.set_threadpool_model')
    @mock.patch('glance.common.wsgi_app._get_config_files')
    def test_wsgi_init_registers_exit_handler(self, mock_config_files,
                                              mock_set_model,
                                              mock_load, mock_exit):
        mock_config_files.return_value = []
        wsgi_app.init_app()
        mock_exit.assert_called_once_with(wsgi_app.drain_threadpools)

    @mock.patch('glance.async_._THREADPOOL_MODEL', new=None)
    def test_drain_threadpools(self):
        # Initialize the thread pool model and tasks_pool, like API
        # under WSGI would, and so we have a pointer to that exact
        # pool object in the cache
        glance.async_.set_threadpool_model('native')
        model = common.get_thread_pool('tasks_pool')

        with mock.patch.object(model.pool, 'shutdown') as mock_shutdown:
            wsgi_app.drain_threadpools()
            # Make sure that shutdown() was called on the tasks_pool
            # ThreadPoolExecutor
            mock_shutdown.assert_called_once_with()

    @mock.patch('glance.common.config.load_paste_app')
    @mock.patch('glance.async_.set_threadpool_model')
    @mock.patch('glance.common.wsgi_app._get_config_files')
    def test_policy_enforcement_kills_service_if_misconfigured(
            self, mock_load_app, mock_set, mock_config_files):
        self.config(enforce_new_defaults=True, group='oslo_policy')
        self.config(enforce_secure_rbac=False)
        self.assertRaises(exception.ServerError, wsgi_app.init_app)

        self.config(enforce_new_defaults=False, group='oslo_policy')
        self.config(enforce_secure_rbac=True)
        self.assertRaises(exception.ServerError, wsgi_app.init_app)

    @mock.patch('glance.common.config.load_paste_app')
    @mock.patch('glance.async_.set_threadpool_model')
    @mock.patch('glance.common.wsgi_app._get_config_files')
    def test_policy_enforcement_valid_truthy_configuration(
            self, mock_load_app, mock_set, mock_config_files):
        self.config(enforce_new_defaults=True, group='oslo_policy')
        self.config(enforce_secure_rbac=True)
        self.assertTrue(wsgi_app.init_app())

    @mock.patch('glance.common.config.load_paste_app')
    @mock.patch('glance.async_.set_threadpool_model')
    @mock.patch('glance.common.wsgi_app._get_config_files')
    def test_policy_enforcement_valid_falsy_configuration(
            self, mock_load_app, mock_set, mock_config_files):
        # This is effectively testing the default values, but we're doing that
        # to make sure nothing bad happens at runtime in the default case when
        # validating policy enforcement configuration.
        self.config(enforce_new_defaults=False, group='oslo_policy')
        self.config(enforce_secure_rbac=False)
        self.assertTrue(wsgi_app.init_app())

    @mock.patch('glance.async_._THREADPOOL_MODEL', new=None)
    @mock.patch('glance.common.config.load_paste_app')
    @mock.patch('glance.common.wsgi_app._get_config_files')
    @mock.patch('threading.Thread')
    @mock.patch('glance.housekeeping.StagingStoreCleaner')
    def test_runs_staging_cleanup(self, mock_cleaner, mock_Thread, mock_conf,
                                  mock_load):
        mock_conf.return_value = []
        wsgi_app.init_app()
        mock_Thread.assert_called_once_with(
            target=mock_cleaner().clean_orphaned_staging_residue,
            daemon=True)
        mock_Thread.return_value.start.assert_called_once_with()



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import futurist
from oslo_log import log as logging

from glance.i18n import _LE


LOG = logging.getLogger(__name__)


class TaskExecutor(object):
    """Base class for Asynchronous task executors. It does not support the
    execution mechanism.

    Provisions the extensible classes with necessary variables to utilize
    important Glance modules like, context, task_repo, image_repo,
    image_factory.

    Note:
        It also gives abstraction for the standard pre-processing and
        post-processing operations to be executed by a task. These may include
        validation checks, security checks, introspection, error handling etc.
        The aim is to give developers an abstract sense of the execution
        pipeline logic.

    Args:
        context: glance.context.RequestContext object for AuthZ and AuthN
            checks
        task_repo: glance.db.TaskRepo object which acts as a translator for
            glance.domain.Task and glance.domain.TaskStub objects
            into ORM semantics
        image_repo: glance.db.ImageRepo object which acts as a translator for
            glance.domain.Image object into ORM semantics
        image_factory: glance.domain.ImageFactory object to be used for
            creating new images for certain types of tasks viz. import, cloning
        admin_repo: glance.db.ImageRepo object which acts as a translator for
            glance.domain.Image object into ORM semantics, but with an admin
            context (optional)
    """

    def __init__(self, context, task_repo, image_repo, image_factory,
                 admin_repo=None):
        self.context = context
        self.task_repo = task_repo
        self.image_repo = image_repo
        self.image_factory = image_factory
        self.admin_repo = admin_repo

    def begin_processing(self, task_id):
        task = self.task_repo.get(task_id)
        task.begin_processing()
        self.task_repo.save(task)

        # start running
        self._run(task_id, task.type)

    def _run(self, task_id, task_type):
        task = self.task_repo.get(task_id)
        msg = _LE("This execution of Tasks is not setup. Please consult the "
                  "project documentation for more information on the "
                  "executors available.")
        LOG.error(msg)
        task.fail(_LE("Internal error occurred while trying to process task."))
        self.task_repo.save(task)


class ThreadPoolModel(object):
    """Base class for an abstract ThreadPool.

    Do not instantiate this directly, use one of the concrete
    implementations.
    """

    DEFAULTSIZE = 1

    @staticmethod
    def get_threadpool_executor_class():
        """Returns a futurist.ThreadPoolExecutor class."""
        pass

    def __init__(self, size=None):
        if size is None:
            size = self.DEFAULTSIZE

        threadpool_cls = self.get_threadpool_executor_class()
        LOG.debug('Creating threadpool model %r with size %i',
                  threadpool_cls.__name__, size)
        self.pool = threadpool_cls(size)

    def spawn(self, fn, *args, **kwargs):
        """Spawn a function with args using the thread pool."""
        LOG.debug('Spawning with %s: %s(%s, %s)' % (
            self.get_threadpool_executor_class().__name__,
            fn, args, kwargs))
        return self.pool.submit(fn, *args, **kwargs)


class EventletThreadPoolModel(ThreadPoolModel):
    """A ThreadPoolModel suitable for use with evenlet/greenthreads."""
    DEFAULTSIZE = 1024

    @staticmethod
    def get_threadpool_executor_class():
        return futurist.GreenThreadPoolExecutor


class NativeThreadPoolModel(ThreadPoolModel):
    """A ThreadPoolModel suitable for use with native threads."""
    DEFAULTSIZE = 16

    @staticmethod
    def get_threadpool_executor_class():
        return futurist.ThreadPoolExecutor


_THREADPOOL_MODEL = None


def set_threadpool_model(thread_type):
    """Set the system-wide threadpool model.

    This sets the type of ThreadPoolModel to use globally in the process.
    It should be called very early in init, and only once.

    :param thread_type: A string indicating the threading type in use,
                        either "eventlet" or "native"
    :raises: RuntimeError if the model is already set or some thread_type
             other than one of the supported ones is provided.
    """
    global _THREADPOOL_MODEL

    if thread_type == 'native':
        model = NativeThreadPoolModel
    elif thread_type == 'eventlet':
        model = EventletThreadPoolModel
    else:
        raise RuntimeError(
            ('Invalid thread type %r '
             '(must be "native" or "eventlet")') % (thread_type))

    if _THREADPOOL_MODEL is model:
        # Re-setting the same model is fine...
        return

    if _THREADPOOL_MODEL is not None:
        # ...changing it is not.
        raise RuntimeError('Thread model is already set')

    LOG.info('Threadpool model set to %r', model.__name__)
    _THREADPOOL_MODEL = model


def get_threadpool_model():
    """Returns the system-wide threadpool model class.

    This must be called after set_threadpool_model() whenever
    some code needs to know what the threadpool implementation is.

    This may only be called after set_threadpool_model() has been
    called to set the desired threading mode. If it is called before
    the model is set, it will raise AssertionError. This would likely
    be the case if this got run in a test before the model was
    initialized, or if glance modules that use threading were imported
    and run from some other code without setting the model first.

    :raises: AssertionError if the model has not yet been set.
    """
    global _THREADPOOL_MODEL
    assert _THREADPOOL_MODEL
    return _THREADPOOL_MODEL



# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils
import requests
from six.moves import http_client as http

from glance.tests import functional


class TestSchemas(functional.FunctionalTest):

    def setUp(self):
        super(TestSchemas, self).setUp()
        self.cleanup()
        self.start_servers(**self.__dict__.copy())

    def test_resource(self):
        # Ensure the image link works and custom properties are loaded
        path = 'http://%s:%d/v2/schemas/image' % ('127.0.0.1', self.api_port)
        response = requests.get(path)
        self.assertEqual(http.OK, response.status_code)
        image_schema = jsonutils.loads(response.text)
        expected = set([
            'id',
            'name',
            'visibility',
            'checksum',
            'os_hash_algo',
            'os_hash_value',
            'created_at',
            'updated_at',
            'tags',
            'size',
            'virtual_size',
            'owner',
            'container_format',
            'disk_format',
            'self',
            'file',
            'status',
            'schema',
            'direct_url',
            'locations',
            'min_ram',
            'min_disk',
            'protected',
            'os_hidden',
            'stores'
        ])
        self.assertEqual(expected, set(image_schema['properties'].keys()))

        # Ensure the images link works and agrees with the image schema
        path = 'http://%s:%d/v2/schemas/images' % ('127.0.0.1', self.api_port)
        response = requests.get(path)
        self.assertEqual(http.OK, response.status_code)
        images_schema = jsonutils.loads(response.text)
        item_schema = images_schema['properties']['images']['items']
        self.assertEqual(item_schema, image_schema)

        self.stop_servers()



# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
# Copyright 2013 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy import MetaData

tables = ['image_locations']


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine
    if migrate_engine.name == "mysql":
        d = migrate_engine.execute("SHOW TABLE STATUS WHERE Engine!='InnoDB';")
        for row in d.fetchall():
            table_name = row[0]
            if table_name in tables:
                migrate_engine.execute("ALTER TABLE %s Engine=InnoDB" %
                                       table_name)



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from unittest import mock

import futurist
import glance_store as store
from oslo_config import cfg
from taskflow.patterns import linear_flow

import glance.async_
from glance.async_.flows import api_image_import
import glance.tests.utils as test_utils

CONF = cfg.CONF


class TestTaskExecutor(test_utils.BaseTestCase):

    def setUp(self):
        super(TestTaskExecutor, self).setUp()
        self.context = mock.Mock()
        self.task_repo = mock.Mock()
        self.image_repo = mock.Mock()
        self.image_factory = mock.Mock()
        self.executor = glance.async_.TaskExecutor(self.context,
                                                   self.task_repo,
                                                   self.image_repo,
                                                   self.image_factory)

    def test_begin_processing(self):
        # setup
        task_id = mock.ANY
        task_type = mock.ANY
        task = mock.Mock()

        with mock.patch.object(
                glance.async_.TaskExecutor,
                '_run') as mock_run:
            self.task_repo.get.return_value = task
            self.executor.begin_processing(task_id)

        # assert the call
        mock_run.assert_called_once_with(task_id, task_type)

    def test_with_admin_repo(self):
        admin_repo = mock.MagicMock()
        executor = glance.async_.TaskExecutor(self.context,
                                              self.task_repo,
                                              self.image_repo,
                                              self.image_factory,
                                              admin_repo=admin_repo)
        self.assertEqual(admin_repo, executor.admin_repo)


class TestImportTaskFlow(test_utils.BaseTestCase):

    def setUp(self):
        super(TestImportTaskFlow, self).setUp()
        store.register_opts(CONF)
        self.config(default_store='file',
                    stores=['file', 'http'],
                    filesystem_store_datadir=self.test_dir,
                    group="glance_store")
        self.config(enabled_import_methods=[
            'glance-direct', 'web-download', 'copy-image'])
        self.config(node_staging_uri='file:///tmp/staging')
        store.create_stores(CONF)
        self.base_flow = ['ImageLock', 'ConfigureStaging', 'ImportToStore',
                          'DeleteFromFS', 'VerifyImageState',
                          'CompleteTask']
        self.import_plugins = ['Convert_Image',
                               'Decompress_Image',
                               'InjectMetadataProperties']

    def _get_flow(self, import_req=None):
        inputs = {
            'task_id': mock.sentinel.task_id,
            'task_type': mock.MagicMock(),
            'task_repo': mock.MagicMock(),
            'image_repo': mock.MagicMock(),
            'image_id': mock.MagicMock(),
            'import_req': import_req or mock.MagicMock()
        }
        inputs['image_repo'].get.return_value = mock.MagicMock(
            extra_properties={'os_glance_import_task': mock.sentinel.task_id})
        flow = api_image_import.get_flow(**inputs)
        return flow

    def _get_flow_tasks(self, flow):
        flow_comp = []
        for c, p in flow.iter_nodes():
            if isinstance(c, linear_flow.Flow):
                flow_comp += self._get_flow_tasks(c)
            else:
                name = str(c).split('-')
                if len(name) > 1:
                    flow_comp.append(name[1])
        return flow_comp

    def test_get_default_flow(self):
        # This test will ensure that without import plugins
        # and without internal plugins flow builds with the
        # base_flow components
        flow = self._get_flow()

        flow_comp = self._get_flow_tasks(flow)
        # assert flow has all the tasks
        self.assertEqual(len(self.base_flow), len(flow_comp))
        for c in self.base_flow:
            self.assertIn(c, flow_comp)

    def test_get_flow_web_download_enabled(self):
        # This test will ensure that without import plugins
        # and with web-download plugin flow builds with
        # base_flow components and '_WebDownload'
        import_req = {
            'method': {
                'name': 'web-download',
                'uri': 'http://cloud.foo/image.qcow2'
            }
        }

        flow = self._get_flow(import_req=import_req)

        flow_comp = self._get_flow_tasks(flow)
        # assert flow has all the tasks
        self.assertEqual(len(self.base_flow) + 1, len(flow_comp))
        for c in self.base_flow:
            self.assertIn(c, flow_comp)
        self.assertIn('WebDownload', flow_comp)

    @mock.patch.object(store, 'get_store_from_store_identifier')
    def test_get_flow_copy_image_enabled(self, mock_store):
        # This test will ensure that without import plugins
        # and with copy-image plugin flow builds with
        # base_flow components and '_CopyImage'
        import_req = {
            'method': {
                'name': 'copy-image',
                'stores': ['fake-store']
            }
        }

        mock_store.return_value = mock.Mock()
        flow = self._get_flow(import_req=import_req)

        flow_comp = self._get_flow_tasks(flow)
        # assert flow has all the tasks
        self.assertEqual(len(self.base_flow) + 1, len(flow_comp))
        for c in self.base_flow:
            self.assertIn(c, flow_comp)
        self.assertIn('CopyImage', flow_comp)

    def test_get_flow_with_all_plugins_enabled(self):
        # This test will ensure that flow includes import plugins
        # and base flow
        self.config(image_import_plugins=['image_conversion',
                                          'image_decompression',
                                          'inject_image_metadata'],
                    group='image_import_opts')

        flow = self._get_flow()

        flow_comp = self._get_flow_tasks(flow)
        # assert flow has all the tasks (base_flow + plugins)
        plugins = CONF.image_import_opts.image_import_plugins
        self.assertEqual(len(self.base_flow) + len(plugins), len(flow_comp))
        for c in self.base_flow:
            self.assertIn(c, flow_comp)
        for c in self.import_plugins:
            self.assertIn(c, flow_comp)

    @mock.patch.object(store, 'get_store_from_store_identifier')
    def test_get_flow_copy_image_not_includes_import_plugins(
            self, mock_store):
        # This test will ensure that flow does not includes import
        # plugins as import method is copy image
        self.config(image_import_plugins=['image_conversion',
                                          'image_decompression',
                                          'inject_image_metadata'],
                    group='image_import_opts')

        mock_store.return_value = mock.Mock()
        import_req = {
            'method': {
                'name': 'copy-image',
                'stores': ['fake-store']
            }
        }

        flow = self._get_flow(import_req=import_req)

        flow_comp = self._get_flow_tasks(flow)
        # assert flow has all the tasks (just base and conversion)
        self.assertEqual(len(self.base_flow) + 1, len(flow_comp))
        for c in self.base_flow:
            self.assertIn(c, flow_comp)
        self.assertIn('CopyImage', flow_comp)


@mock.patch('glance.async_._THREADPOOL_MODEL', new=None)
class TestSystemThreadPoolModel(test_utils.BaseTestCase):
    def test_eventlet_model(self):
        model_cls = glance.async_.EventletThreadPoolModel
        self.assertEqual(futurist.GreenThreadPoolExecutor,
                         model_cls.get_threadpool_executor_class())

    def test_native_model(self):
        model_cls = glance.async_.NativeThreadPoolModel
        self.assertEqual(futurist.ThreadPoolExecutor,
                         model_cls.get_threadpool_executor_class())

    @mock.patch('glance.async_.ThreadPoolModel.get_threadpool_executor_class')
    def test_base_model_spawn(self, mock_gte):
        pool_cls = mock.MagicMock()
        pool_cls.configure_mock(__name__='fake')
        mock_gte.return_value = pool_cls

        model = glance.async_.ThreadPoolModel()
        result = model.spawn(print, 'foo', bar='baz')

        pool = pool_cls.return_value

        # Make sure the default size was passed to the executor
        pool_cls.assert_called_once_with(1)

        # Make sure we submitted the function to the executor
        pool.submit.assert_called_once_with(print, 'foo', bar='baz')

        # This isn't used anywhere, but make sure we get the future
        self.assertEqual(pool.submit.return_value, result)

    @mock.patch('glance.async_.ThreadPoolModel.get_threadpool_executor_class')
    def test_base_model_init_with_size(self, mock_gte):
        mock_gte.return_value.__name__ = 'TestModel'
        with mock.patch.object(glance.async_, 'LOG') as mock_log:
            glance.async_.ThreadPoolModel(123)
            mock_log.debug.assert_called_once_with(
                'Creating threadpool model %r with size %i',
                'TestModel', 123)
        mock_gte.return_value.assert_called_once_with(123)

    def test_set_threadpool_model_native(self):
        glance.async_.set_threadpool_model('native')
        self.assertEqual(glance.async_.NativeThreadPoolModel,
                         glance.async_._THREADPOOL_MODEL)

    def test_set_threadpool_model_eventlet(self):
        glance.async_.set_threadpool_model('eventlet')
        self.assertEqual(glance.async_.EventletThreadPoolModel,
                         glance.async_._THREADPOOL_MODEL)

    def test_set_threadpool_model_unknown(self):
        # Unknown threadpool models are not tolerated
        self.assertRaises(RuntimeError,
                          glance.async_.set_threadpool_model,
                          'danthread9000')

    def test_set_threadpool_model_again(self):
        # Setting the model to the same thing is fine
        glance.async_.set_threadpool_model('native')
        glance.async_.set_threadpool_model('native')

    def test_set_threadpool_model_different(self):
        glance.async_.set_threadpool_model('native')
        # The model cannot be switched at runtime
        self.assertRaises(RuntimeError,
                          glance.async_.set_threadpool_model,
                          'eventlet')

    def test_set_threadpool_model_log(self):
        with mock.patch.object(glance.async_, 'LOG') as mock_log:
            glance.async_.set_threadpool_model('eventlet')
            mock_log.info.assert_called_once_with(
                'Threadpool model set to %r', 'EventletThreadPoolModel')

    def test_get_threadpool_model(self):
        glance.async_.set_threadpool_model('native')
        self.assertEqual(glance.async_.NativeThreadPoolModel,
                         glance.async_.get_threadpool_model())

    def test_get_threadpool_model_unset(self):
        # If the model is not set, we get an AssertionError
        self.assertRaises(AssertionError,
                          glance.async_.get_threadpool_model)



# Copyright 2013 OpenStack Foundation
# Copyright 2013 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


def _proxy(target, attr):
    def get_attr(self):
        return getattr(getattr(self, target), attr)

    def set_attr(self, value):
        return setattr(getattr(self, target), attr, value)

    def del_attr(self):
        return delattr(getattr(self, target), attr)

    return property(get_attr, set_attr, del_attr)


class Helper(object):
    def __init__(self, proxy_class=None, proxy_kwargs=None):
        self.proxy_class = proxy_class
        self.proxy_kwargs = proxy_kwargs or {}

    def proxy(self, obj):
        if obj is None or self.proxy_class is None:
            return obj
        return self.proxy_class(obj, **self.proxy_kwargs)

    def unproxy(self, obj):
        if obj is None or self.proxy_class is None:
            return obj
        return obj.base


class TaskRepo(object):
    def __init__(self, base,
                 task_proxy_class=None, task_proxy_kwargs=None):
        self.base = base
        self.task_proxy_helper = Helper(task_proxy_class, task_proxy_kwargs)

    def get(self, task_id):
        task = self.base.get(task_id)
        return self.task_proxy_helper.proxy(task)

    def add(self, task):
        self.base.add(self.task_proxy_helper.unproxy(task))

    def save(self, task):
        self.base.save(self.task_proxy_helper.unproxy(task))

    def remove(self, task):
        base_task = self.task_proxy_helper.unproxy(task)
        self.base.remove(base_task)


class TaskStubRepo(object):
    def __init__(self, base, task_stub_proxy_class=None,
                 task_stub_proxy_kwargs=None):
        self.base = base
        self.task_stub_proxy_helper = Helper(task_stub_proxy_class,
                                             task_stub_proxy_kwargs)

    def list(self, *args, **kwargs):
        tasks = self.base.list(*args, **kwargs)
        return [self.task_stub_proxy_helper.proxy(task) for task in tasks]


class Repo(object):
    def __init__(self, base, item_proxy_class=None, item_proxy_kwargs=None):
        self.base = base
        self.helper = Helper(item_proxy_class, item_proxy_kwargs)

    def get(self, item_id):
        return self.helper.proxy(self.base.get(item_id))

    def list(self, *args, **kwargs):
        items = self.base.list(*args, **kwargs)
        return [self.helper.proxy(item) for item in items]

    def add(self, item):
        base_item = self.helper.unproxy(item)
        result = self.base.add(base_item)
        return self.helper.proxy(result)

    def save(self, item, from_state=None):
        base_item = self.helper.unproxy(item)
        result = self.base.save(base_item, from_state=from_state)
        return self.helper.proxy(result)

    def remove(self, item):
        base_item = self.helper.unproxy(item)
        result = self.base.remove(base_item)
        return self.helper.proxy(result)

    def set_property_atomic(self, item, name, value):
        msg = '%s is only valid for images' % __name__
        assert hasattr(item, 'image_id'), msg
        self.base.set_property_atomic(item, name, value)

    def delete_property_atomic(self, item, name, value):
        msg = '%s is only valid for images' % __name__
        assert hasattr(item, 'image_id'), msg
        self.base.delete_property_atomic(item, name, value)


class MemberRepo(object):
    def __init__(self, image, base,
                 member_proxy_class=None, member_proxy_kwargs=None):
        self.image = image
        self.base = base
        self.member_proxy_helper = Helper(member_proxy_class,
                                          member_proxy_kwargs)

    def get(self, member_id):
        member = self.base.get(member_id)
        return self.member_proxy_helper.proxy(member)

    def add(self, member):
        self.base.add(self.member_proxy_helper.unproxy(member))

    def list(self, *args, **kwargs):
        members = self.base.list(*args, **kwargs)
        return [self.member_proxy_helper.proxy(member) for member
                in members]

    def remove(self, member):
        base_item = self.member_proxy_helper.unproxy(member)
        result = self.base.remove(base_item)
        return self.member_proxy_helper.proxy(result)

    def save(self, member, from_state=None):
        base_item = self.member_proxy_helper.unproxy(member)
        result = self.base.save(base_item, from_state=from_state)
        return self.member_proxy_helper.proxy(result)


class ImageFactory(object):
    def __init__(self, base, proxy_class=None, proxy_kwargs=None):
        self.helper = Helper(proxy_class, proxy_kwargs)
        self.base = base

    def new_image(self, **kwargs):
        return self.helper.proxy(self.base.new_image(**kwargs))


class ImageMembershipFactory(object):
    def __init__(self, base, proxy_class=None, proxy_kwargs=None):
        self.helper = Helper(proxy_class, proxy_kwargs)
        self.base = base

    def new_image_member(self, image, member, **kwargs):
        return self.helper.proxy(self.base.new_image_member(image,
                                                            member,
                                                            **kwargs))


class Image(object):
    def __init__(self, base, member_repo_proxy_class=None,
                 member_repo_proxy_kwargs=None):
        self.base = base
        self.helper = Helper(member_repo_proxy_class,
                             member_repo_proxy_kwargs)

    name = _proxy('base', 'name')
    image_id = _proxy('base', 'image_id')
    status = _proxy('base', 'status')
    created_at = _proxy('base', 'created_at')
    updated_at = _proxy('base', 'updated_at')
    visibility = _proxy('base', 'visibility')
    min_disk = _proxy('base', 'min_disk')
    min_ram = _proxy('base', 'min_ram')
    protected = _proxy('base', 'protected')
    os_hidden = _proxy('base', 'os_hidden')
    locations = _proxy('base', 'locations')
    checksum = _proxy('base', 'checksum')
    os_hash_algo = _proxy('base', 'os_hash_algo')
    os_hash_value = _proxy('base', 'os_hash_value')
    owner = _proxy('base', 'owner')
    disk_format = _proxy('base', 'disk_format')
    container_format = _proxy('base', 'container_format')
    size = _proxy('base', 'size')
    virtual_size = _proxy('base', 'virtual_size')
    extra_properties = _proxy('base', 'extra_properties')
    tags = _proxy('base', 'tags')

    def delete(self):
        self.base.delete()

    def deactivate(self):
        self.base.deactivate()

    def reactivate(self):
        self.base.reactivate()

    def set_data(self, data, size=None, backend=None, set_active=True):
        self.base.set_data(data, size, backend=backend, set_active=set_active)

    def get_data(self, *args, **kwargs):
        return self.base.get_data(*args, **kwargs)


class ImageMember(object):
    def __init__(self, base):
        self.base = base

    id = _proxy('base', 'id')
    image_id = _proxy('base', 'image_id')
    member_id = _proxy('base', 'member_id')
    status = _proxy('base', 'status')
    created_at = _proxy('base', 'created_at')
    updated_at = _proxy('base', 'updated_at')


class Task(object):
    def __init__(self, base):
        self.base = base

    task_id = _proxy('base', 'task_id')
    type = _proxy('base', 'type')
    status = _proxy('base', 'status')
    owner = _proxy('base', 'owner')
    expires_at = _proxy('base', 'expires_at')
    created_at = _proxy('base', 'created_at')
    updated_at = _proxy('base', 'updated_at')
    task_input = _proxy('base', 'task_input')
    result = _proxy('base', 'result')
    message = _proxy('base', 'message')
    image_id = _proxy('base', 'image_id')
    request_id = _proxy('base', 'request_id')
    user_id = _proxy('base', 'user_id')

    def begin_processing(self):
        self.base.begin_processing()

    def succeed(self, result):
        self.base.succeed(result)

    def fail(self, message):
        self.base.fail(message)

    def run(self, executor):
        self.base.run(executor)


class TaskStub(object):
    def __init__(self, base):
        self.base = base

    task_id = _proxy('base', 'task_id')
    type = _proxy('base', 'type')
    status = _proxy('base', 'status')
    owner = _proxy('base', 'owner')
    expires_at = _proxy('base', 'expires_at')
    created_at = _proxy('base', 'created_at')
    updated_at = _proxy('base', 'updated_at')


class TaskFactory(object):
    def __init__(self,
                 base,
                 task_proxy_class=None,
                 task_proxy_kwargs=None):
        self.task_helper = Helper(task_proxy_class, task_proxy_kwargs)
        self.base = base

    def new_task(self, **kwargs):
        t = self.base.new_task(**kwargs)
        return self.task_helper.proxy(t)


# Metadef Namespace classes
class MetadefNamespaceRepo(object):
    def __init__(self, base,
                 namespace_proxy_class=None, namespace_proxy_kwargs=None):
        self.base = base
        self.namespace_proxy_helper = Helper(namespace_proxy_class,
                                             namespace_proxy_kwargs)

    def get(self, namespace):
        namespace_obj = self.base.get(namespace)
        return self.namespace_proxy_helper.proxy(namespace_obj)

    def add(self, namespace):
        self.base.add(self.namespace_proxy_helper.unproxy(namespace))

    def list(self, *args, **kwargs):
        namespaces = self.base.list(*args, **kwargs)
        return [self.namespace_proxy_helper.proxy(namespace) for namespace
                in namespaces]

    def remove(self, item):
        base_item = self.namespace_proxy_helper.unproxy(item)
        result = self.base.remove(base_item)
        return self.namespace_proxy_helper.proxy(result)

    def remove_objects(self, item):
        base_item = self.namespace_proxy_helper.unproxy(item)
        result = self.base.remove_objects(base_item)
        return self.namespace_proxy_helper.proxy(result)

    def remove_properties(self, item):
        base_item = self.namespace_proxy_helper.unproxy(item)
        result = self.base.remove_properties(base_item)
        return self.namespace_proxy_helper.proxy(result)

    def remove_tags(self, item):
        base_item = self.namespace_proxy_helper.unproxy(item)
        result = self.base.remove_tags(base_item)
        return self.namespace_proxy_helper.proxy(result)

    def save(self, item):
        base_item = self.namespace_proxy_helper.unproxy(item)
        result = self.base.save(base_item)
        return self.namespace_proxy_helper.proxy(result)


class MetadefNamespace(object):
    def __init__(self, base):
        self.base = base

    namespace_id = _proxy('base', 'namespace_id')
    namespace = _proxy('base', 'namespace')
    display_name = _proxy('base', 'display_name')
    description = _proxy('base', 'description')
    owner = _proxy('base', 'owner')
    visibility = _proxy('base', 'visibility')
    protected = _proxy('base', 'protected')
    created_at = _proxy('base', 'created_at')
    updated_at = _proxy('base', 'updated_at')

    def delete(self):
        self.base.delete()


class MetadefNamespaceFactory(object):
    def __init__(self,
                 base,
                 meta_namespace_proxy_class=None,
                 meta_namespace_proxy_kwargs=None):
        self.meta_namespace_helper = Helper(meta_namespace_proxy_class,
                                            meta_namespace_proxy_kwargs)
        self.base = base

    def new_namespace(self, **kwargs):
        t = self.base.new_namespace(**kwargs)
        return self.meta_namespace_helper.proxy(t)


# Metadef object classes
class MetadefObjectRepo(object):
    def __init__(self, base,
                 object_proxy_class=None, object_proxy_kwargs=None):
        self.base = base
        self.object_proxy_helper = Helper(object_proxy_class,
                                          object_proxy_kwargs)

    def get(self, namespace, object_name):
        meta_object = self.base.get(namespace, object_name)
        return self.object_proxy_helper.proxy(meta_object)

    def add(self, meta_object):
        self.base.add(self.object_proxy_helper.unproxy(meta_object))

    def list(self, *args, **kwargs):
        objects = self.base.list(*args, **kwargs)
        return [self.object_proxy_helper.proxy(meta_object) for meta_object
                in objects]

    def remove(self, item):
        base_item = self.object_proxy_helper.unproxy(item)
        result = self.base.remove(base_item)
        return self.object_proxy_helper.proxy(result)

    def save(self, item):
        base_item = self.object_proxy_helper.unproxy(item)
        result = self.base.save(base_item)
        return self.object_proxy_helper.proxy(result)


class MetadefObject(object):
    def __init__(self, base):
        self.base = base
    namespace = _proxy('base', 'namespace')
    object_id = _proxy('base', 'object_id')
    name = _proxy('base', 'name')
    required = _proxy('base', 'required')
    description = _proxy('base', 'description')
    properties = _proxy('base', 'properties')
    created_at = _proxy('base', 'created_at')
    updated_at = _proxy('base', 'updated_at')

    def delete(self):
        self.base.delete()


class MetadefObjectFactory(object):
    def __init__(self,
                 base,
                 meta_object_proxy_class=None,
                 meta_object_proxy_kwargs=None):
        self.meta_object_helper = Helper(meta_object_proxy_class,
                                         meta_object_proxy_kwargs)
        self.base = base

    def new_object(self, **kwargs):
        t = self.base.new_object(**kwargs)
        return self.meta_object_helper.proxy(t)


# Metadef ResourceType classes
class MetadefResourceTypeRepo(object):
    def __init__(self, base, resource_type_proxy_class=None,
                 resource_type_proxy_kwargs=None):
        self.base = base
        self.resource_type_proxy_helper = Helper(resource_type_proxy_class,
                                                 resource_type_proxy_kwargs)

    def add(self, meta_resource_type):
        self.base.add(self.resource_type_proxy_helper.unproxy(
            meta_resource_type))

    def get(self, *args, **kwargs):
        resource_type = self.base.get(*args, **kwargs)
        return self.resource_type_proxy_helper.proxy(resource_type)

    def list(self, *args, **kwargs):
        resource_types = self.base.list(*args, **kwargs)
        return [self.resource_type_proxy_helper.proxy(resource_type)
                for resource_type in resource_types]

    def remove(self, item):
        base_item = self.resource_type_proxy_helper.unproxy(item)
        result = self.base.remove(base_item)
        return self.resource_type_proxy_helper.proxy(result)


class MetadefResourceType(object):
    def __init__(self, base):
        self.base = base
    namespace = _proxy('base', 'namespace')
    name = _proxy('base', 'name')
    prefix = _proxy('base', 'prefix')
    properties_target = _proxy('base', 'properties_target')
    created_at = _proxy('base', 'created_at')
    updated_at = _proxy('base', 'updated_at')

    def delete(self):
        self.base.delete()


class MetadefResourceTypeFactory(object):
    def __init__(self,
                 base,
                 resource_type_proxy_class=None,
                 resource_type_proxy_kwargs=None):
        self.resource_type_helper = Helper(resource_type_proxy_class,
                                           resource_type_proxy_kwargs)
        self.base = base

    def new_resource_type(self, **kwargs):
        t = self.base.new_resource_type(**kwargs)
        return self.resource_type_helper.proxy(t)


# Metadef namespace property classes
class MetadefPropertyRepo(object):
    def __init__(self, base,
                 property_proxy_class=None, property_proxy_kwargs=None):
        self.base = base
        self.property_proxy_helper = Helper(property_proxy_class,
                                            property_proxy_kwargs)

    def get(self, namespace, property_name):
        property = self.base.get(namespace, property_name)
        return self.property_proxy_helper.proxy(property)

    def add(self, property):
        self.base.add(self.property_proxy_helper.unproxy(property))

    def list(self, *args, **kwargs):
        properties = self.base.list(*args, **kwargs)
        return [self.property_proxy_helper.proxy(property) for property
                in properties]

    def remove(self, item):
        base_item = self.property_proxy_helper.unproxy(item)
        result = self.base.remove(base_item)
        return self.property_proxy_helper.proxy(result)

    def save(self, item):
        base_item = self.property_proxy_helper.unproxy(item)
        result = self.base.save(base_item)
        return self.property_proxy_helper.proxy(result)


class MetadefProperty(object):
    def __init__(self, base):
        self.base = base
    namespace = _proxy('base', 'namespace')
    property_id = _proxy('base', 'property_id')
    name = _proxy('base', 'name')
    schema = _proxy('base', 'schema')

    def delete(self):
        self.base.delete()


class MetadefPropertyFactory(object):
    def __init__(self,
                 base,
                 property_proxy_class=None,
                 property_proxy_kwargs=None):
        self.meta_object_helper = Helper(property_proxy_class,
                                         property_proxy_kwargs)
        self.base = base

    def new_namespace_property(self, **kwargs):
        t = self.base.new_namespace_property(**kwargs)
        return self.meta_object_helper.proxy(t)


# Metadef tag classes
class MetadefTagRepo(object):
    def __init__(self, base,
                 tag_proxy_class=None, tag_proxy_kwargs=None):
        self.base = base
        self.tag_proxy_helper = Helper(tag_proxy_class,
                                       tag_proxy_kwargs)

    def get(self, namespace, name):
        meta_tag = self.base.get(namespace, name)
        return self.tag_proxy_helper.proxy(meta_tag)

    def add(self, meta_tag):
        self.base.add(self.tag_proxy_helper.unproxy(meta_tag))

    def add_tags(self, meta_tags):
        tags_list = []
        for meta_tag in meta_tags:
            tags_list.append(self.tag_proxy_helper.unproxy(meta_tag))
        self.base.add_tags(tags_list)

    def list(self, *args, **kwargs):
        tags = self.base.list(*args, **kwargs)
        return [self.tag_proxy_helper.proxy(meta_tag) for meta_tag
                in tags]

    def remove(self, item):
        base_item = self.tag_proxy_helper.unproxy(item)
        result = self.base.remove(base_item)
        return self.tag_proxy_helper.proxy(result)

    def save(self, item):
        base_item = self.tag_proxy_helper.unproxy(item)
        result = self.base.save(base_item)
        return self.tag_proxy_helper.proxy(result)


class MetadefTag(object):
    def __init__(self, base):
        self.base = base

    namespace = _proxy('base', 'namespace')
    tag_id = _proxy('base', 'tag_id')
    name = _proxy('base', 'name')
    created_at = _proxy('base', 'created_at')
    updated_at = _proxy('base', 'updated_at')

    def delete(self):
        self.base.delete()


class MetadefTagFactory(object):
    def __init__(self,
                 base,
                 meta_tag_proxy_class=None,
                 meta_tag_proxy_kwargs=None):
        self.meta_tag_helper = Helper(meta_tag_proxy_class,
                                      meta_tag_proxy_kwargs)
        self.base = base

    def new_tag(self, **kwargs):
        t = self.base.new_tag(**kwargs)
        return self.meta_tag_helper.proxy(t)



# Copyright (c) 2010 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Glance documentation build configuration file

import os
import sys

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('../..'))
sys.path.insert(0, os.path.abspath('../../bin'))

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'stevedore.sphinxext',
    'sphinx.ext.viewcode',
    'oslo_config.sphinxext',
    'oslo_config.sphinxconfiggen',
    'oslo_policy.sphinxpolicygen',
    'openstackdocstheme',
    'sphinxcontrib.apidoc',
]

# openstackdocstheme options
openstackdocs_repo_name = 'openstack/glance'
openstackdocs_bug_project = 'glance'
openstackdocs_bug_tag = 'documentation'

# sphinxcontrib.apidoc options
apidoc_module_dir = '../../glance'
apidoc_output_dir = 'contributor/api'
apidoc_excluded_paths = [
    'hacking/*',
    'hacking',
    'tests/*',
    'tests',
    'db/sqlalchemy/*',
    'db/sqlalchemy']
apidoc_separate_modules = True

config_generator_config_file = [
    ('../../etc/oslo-config-generator/glance-api.conf',
     '_static/glance-api'),
    ('../../etc/oslo-config-generator/glance-cache.conf',
     '_static/glance-cache'),
    ('../../etc/oslo-config-generator/glance-manage.conf',
     '_static/glance-manage'),
    ('../../etc/oslo-config-generator/glance-scrubber.conf',
     '_static/glance-scrubber'),
]

policy_generator_config_file = [
    ('../../etc/glance-policy-generator.conf', '_static/glance'),
]

# The master toctree document.
master_doc = 'index'

# General information about the project.
copyright = u'2010-present, OpenStack Foundation.'

exclude_patterns = [
    # The man directory includes some snippet files that are included
    # in other documents during the build but that should not be
    # included in the toctree themselves, so tell Sphinx to ignore
    # them when scanning for input files.
    'cli/footer.txt',
    'cli/general_options.txt',
    'cli/openstack_options.txt',
]

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
show_authors = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['glance.']

# -- Options for man page output --------------------------------------------

# Grouping the document tree for man pages.
# List of tuples 'sourcefile', 'target', u'title', u'Authors name', 'manual'

man_pages = [
    ('cli/glanceapi', 'glance-api', u'Glance API Server',
     [u'OpenStack'], 1),
    ('cli/glancecachecleaner', 'glance-cache-cleaner', u'Glance Cache Cleaner',
     [u'OpenStack'], 1),
    ('cli/glancecachemanage', 'glance-cache-manage', u'Glance Cache Manager',
     [u'OpenStack'], 1),
    ('cli/glancecacheprefetcher', 'glance-cache-prefetcher',
     u'Glance Cache Pre-fetcher', [u'OpenStack'], 1),
    ('cli/glancecachepruner', 'glance-cache-pruner', u'Glance Cache Pruner',
     [u'OpenStack'], 1),
    ('cli/glancecontrol', 'glance-control', u'Glance Daemon Control Helper ',
     [u'OpenStack'], 1),
    ('cli/glancemanage', 'glance-manage', u'Glance Management Utility',
     [u'OpenStack'], 1),
    ('cli/glancereplicator', 'glance-replicator', u'Glance Replicator',
     [u'OpenStack'], 1),
    ('cli/glancescrubber', 'glance-scrubber', u'Glance Scrubber Service',
     [u'OpenStack'], 1)
]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
# html_theme_path = ["."]
# html_theme = '_theme'
html_theme = 'openstackdocs'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add any paths that contain "extra" files, such as .htaccess or
# robots.txt.
html_extra_path = ['_extra']

# If false, no module index is generated.
html_use_modindex = True

# If false, no index is generated.
html_use_index = True


# -- Options for LaTeX output ------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author,
# documentclass [howto/manual]).
latex_documents = [
    ('index', 'Glance.tex', u'Glance Documentation',
     u'Glance Team', 'manual'),
]



# Copyright 2013 IBM Corp.
# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
While SQLAlchemy/sqlalchemy-migrate should abstract this correctly,
there are known issues with these libraries so SQLite and non-SQLite
migrations must be done separately.
"""

import uuid

import migrate
import sqlalchemy


and_ = sqlalchemy.and_
or_ = sqlalchemy.or_


def upgrade(migrate_engine):
    """
    Call the correct dialect-specific upgrade.
    """
    meta = sqlalchemy.MetaData()
    meta.bind = migrate_engine

    t_images = _get_table('images', meta)
    t_image_members = _get_table('image_members', meta)
    t_image_properties = _get_table('image_properties', meta)

    dialect = migrate_engine.url.get_dialect().name
    if dialect == "sqlite":
        _upgrade_sqlite(meta, t_images, t_image_members, t_image_properties)
        _update_all_ids_to_uuids(t_images, t_image_members, t_image_properties)
    elif dialect == "ibm_db_sa":
        _upgrade_db2(meta, t_images, t_image_members, t_image_properties)
        _update_all_ids_to_uuids(t_images, t_image_members, t_image_properties)
        _add_db2_constraints(meta)
    else:
        _upgrade_other(t_images, t_image_members, t_image_properties, dialect)


def _upgrade_sqlite(meta, t_images, t_image_members, t_image_properties):
    """
    Upgrade 011 -> 012 with special SQLite-compatible logic.
    """

    sql_commands = [
        """CREATE TABLE images_backup (
           id VARCHAR(36) NOT NULL,
           name VARCHAR(255),
           size INTEGER,
           status VARCHAR(30) NOT NULL,
           is_public BOOLEAN NOT NULL,
           location TEXT,
           created_at DATETIME NOT NULL,
           updated_at DATETIME,
           deleted_at DATETIME,
           deleted BOOLEAN NOT NULL,
           disk_format VARCHAR(20),
           container_format VARCHAR(20),
           checksum VARCHAR(32),
           owner VARCHAR(255),
           min_disk INTEGER NOT NULL,
           min_ram INTEGER NOT NULL,
           PRIMARY KEY (id),
           CHECK (is_public IN (0, 1)),
           CHECK (deleted IN (0, 1))
        );""",
        """INSERT INTO images_backup
           SELECT * FROM images;""",
        """CREATE TABLE image_members_backup (
            id INTEGER NOT NULL,
            image_id VARCHAR(36) NOT NULL,
            member VARCHAR(255) NOT NULL,
            can_share BOOLEAN NOT NULL,
            created_at DATETIME NOT NULL,
            updated_at DATETIME,
            deleted_at DATETIME,
            deleted BOOLEAN NOT NULL,
            PRIMARY KEY (id),
            UNIQUE (image_id, member),
            CHECK (can_share IN (0, 1)),
            CHECK (deleted IN (0, 1)),
            FOREIGN KEY(image_id) REFERENCES images (id)
        );""",
        """INSERT INTO image_members_backup
            SELECT * FROM image_members;""",
        """CREATE TABLE image_properties_backup (
            id INTEGER NOT NULL,
            image_id VARCHAR(36) NOT NULL,
            name VARCHAR(255) NOT NULL,
            value TEXT,
            created_at DATETIME NOT NULL,
            updated_at DATETIME,
            deleted_at DATETIME,
            deleted BOOLEAN NOT NULL,
            PRIMARY KEY (id),
            CHECK (deleted IN (0, 1)),
            UNIQUE (image_id, name),
            FOREIGN KEY(image_id) REFERENCES images (id)
        );""",
        """INSERT INTO image_properties_backup
            SELECT * FROM image_properties;""",
    ]

    for command in sql_commands:
        meta.bind.execute(command)

    _sqlite_table_swap(meta, t_image_members, t_image_properties, t_images)


def _upgrade_db2(meta, t_images, t_image_members, t_image_properties):
    """
    Upgrade for DB2.
    """
    t_images.c.id.alter(sqlalchemy.String(36), primary_key=True)

    image_members_backup = sqlalchemy.Table(
        'image_members_backup',
        meta,
        sqlalchemy.Column('id',
                          sqlalchemy.Integer(),
                          primary_key=True,
                          nullable=False),
        sqlalchemy.Column('image_id',
                          sqlalchemy.String(36),
                          nullable=False,
                          index=True),
        sqlalchemy.Column('member',
                          sqlalchemy.String(255),
                          nullable=False),
        sqlalchemy.Column('can_share',
                          sqlalchemy.Boolean(),
                          nullable=False,
                          default=False),
        sqlalchemy.Column('created_at',
                          sqlalchemy.DateTime(),
                          nullable=False),
        sqlalchemy.Column('updated_at',
                          sqlalchemy.DateTime()),
        sqlalchemy.Column('deleted_at',
                          sqlalchemy.DateTime()),
        sqlalchemy.Column('deleted',
                          sqlalchemy.Boolean(),
                          nullable=False,
                          default=False,
                          index=True),
        sqlalchemy.UniqueConstraint('image_id', 'member'),
        extend_existing=True)

    image_properties_backup = sqlalchemy.Table(
        'image_properties_backup',
        meta,
        sqlalchemy.Column('id',
                          sqlalchemy.Integer(),
                          primary_key=True,
                          nullable=False),
        sqlalchemy.Column('image_id',
                          sqlalchemy.String(36),
                          nullable=False,
                          index=True),
        sqlalchemy.Column('name',
                          sqlalchemy.String(255),
                          nullable=False),
        sqlalchemy.Column('value',
                          sqlalchemy.Text()),
        sqlalchemy.Column('created_at',
                          sqlalchemy.DateTime(),
                          nullable=False),
        sqlalchemy.Column('updated_at',
                          sqlalchemy.DateTime()),
        sqlalchemy.Column('deleted_at',
                          sqlalchemy.DateTime()),
        sqlalchemy.Column('deleted',
                          sqlalchemy.Boolean(),
                          nullable=False,
                          default=False,
                          index=True),
        sqlalchemy.UniqueConstraint(
            'image_id', 'name',
            name='ix_image_properties_image_id_name'),
        extend_existing=True)

    image_members_backup.create()
    image_properties_backup.create()

    sql_commands = [
        """INSERT INTO image_members_backup
            SELECT * FROM image_members;""",
        """INSERT INTO image_properties_backup
            SELECT * FROM image_properties;""",
    ]

    for command in sql_commands:
        meta.bind.execute(command)

    t_image_members.drop()
    t_image_properties.drop()

    image_members_backup.rename(name='image_members')
    image_properties_backup.rename(name='image_properties')


def _add_db2_constraints(meta):
    # Create the foreign keys
    sql_commands = [
        """ALTER TABLE image_members ADD CONSTRAINT member_image_id
            FOREIGN KEY (image_id)
            REFERENCES images (id);""",
        """ALTER TABLE image_properties ADD CONSTRAINT property_image_id
            FOREIGN KEY (image_id)
            REFERENCES images (id);""",
    ]
    for command in sql_commands:
        meta.bind.execute(command)


def _upgrade_other(t_images, t_image_members, t_image_properties, dialect):
    """
    Upgrade 011 -> 012 with logic for non-SQLite databases.
    """
    foreign_keys = _get_foreign_keys(t_images,
                                     t_image_members,
                                     t_image_properties, dialect)

    for fk in foreign_keys:
        fk.drop()

    t_images.c.id.alter(sqlalchemy.String(36), primary_key=True)
    t_image_members.c.image_id.alter(sqlalchemy.String(36))
    t_image_properties.c.image_id.alter(sqlalchemy.String(36))

    _update_all_ids_to_uuids(t_images, t_image_members, t_image_properties)

    for fk in foreign_keys:
        fk.create()


def _sqlite_table_swap(meta, t_image_members, t_image_properties, t_images):
    t_image_members.drop()
    t_image_properties.drop()
    t_images.drop()

    meta.bind.execute("ALTER TABLE images_backup "
                      "RENAME TO images")
    meta.bind.execute("ALTER TABLE image_members_backup "
                      "RENAME TO image_members")
    meta.bind.execute("ALTER TABLE image_properties_backup "
                      "RENAME TO image_properties")
    meta.bind.execute("""CREATE INDEX ix_image_properties_deleted
                          ON image_properties (deleted);""")
    meta.bind.execute("""CREATE INDEX ix_image_properties_name
                          ON image_properties (name);""")


def _get_table(table_name, metadata):
    """Return a sqlalchemy Table definition with associated metadata."""
    return sqlalchemy.Table(table_name, metadata, autoload=True)


def _get_foreign_keys(t_images, t_image_members, t_image_properties, dialect):
    """Retrieve and return foreign keys for members/properties tables."""
    foreign_keys = []
    if t_image_members.foreign_keys:
        img_members_fk_name = list(t_image_members.foreign_keys)[0].name
        if dialect == 'mysql':
            fk1 = migrate.ForeignKeyConstraint([t_image_members.c.image_id],
                                               [t_images.c.id],
                                               name=img_members_fk_name)
        else:
            fk1 = migrate.ForeignKeyConstraint([t_image_members.c.image_id],
                                               [t_images.c.id])
        foreign_keys.append(fk1)

    if t_image_properties.foreign_keys:
        img_properties_fk_name = list(t_image_properties.foreign_keys)[0].name
        if dialect == 'mysql':
            fk2 = migrate.ForeignKeyConstraint([t_image_properties.c.image_id],
                                               [t_images.c.id],
                                               name=img_properties_fk_name)
        else:
            fk2 = migrate.ForeignKeyConstraint([t_image_properties.c.image_id],
                                               [t_images.c.id])
        foreign_keys.append(fk2)

    return foreign_keys


def _update_all_ids_to_uuids(t_images, t_image_members, t_image_properties):
    """Transition from INTEGER id to VARCHAR(36) id."""
    images = list(t_images.select().execute())

    for image in images:
        old_id = image["id"]
        new_id = str(uuid.uuid4())

        t_images.update().where(
            t_images.c.id == old_id).values(id=new_id).execute()

        t_image_members.update().where(
            t_image_members.c.image_id == old_id).values(
                image_id=new_id).execute()

        t_image_properties.update().where(
            t_image_properties.c.image_id == old_id).values(
                image_id=new_id).execute()

        t_image_properties.update().where(
            and_(or_(t_image_properties.c.name == 'kernel_id',
                     t_image_properties.c.name == 'ramdisk_id'),
                 t_image_properties.c.value == old_id)).values(
                     value=new_id).execute()


def _update_all_uuids_to_ids(t_images, t_image_members, t_image_properties):
    """Transition from VARCHAR(36) id to INTEGER id."""
    images = list(t_images.select().execute())

    new_id = 1
    for image in images:
        old_id = image["id"]

        t_images.update().where(
            t_images.c.id == old_id).values(
                id=str(new_id)).execute()

        t_image_members.update().where(
            t_image_members.c.image_id == old_id).values(
                image_id=str(new_id)).execute()

        t_image_properties.update().where(
            t_image_properties.c.image_id == old_id).values(
                image_id=str(new_id)).execute()

        t_image_properties.update().where(
            and_(or_(t_image_properties.c.name == 'kernel_id',
                     t_image_properties.c.name == 'ramdisk_id'),
                 t_image_properties.c.value == old_id)).values(
                     value=str(new_id)).execute()

        new_id += 1
